/*********************************************************************
 * vic20_kbd.ino - For Arduino Leonardo or compatible boards         *
 * Version 0.1                                                       *
 * vic20_kbd is Copyright (C) 2013 Edgar Puehringer                  *
 * based on Spaceman Spiff's Commodire 64 USB Keyboard (c64key)      *
 * by Mikkel Holm Olsen                                              *
 *********************************************************************
 * vic20_kbd is free software; because it does not (like the         *
 * original) use the V-USB software, there is no need to put it      *
 * under the OBDEV license.                                          *
 *                                                                   *
 * vic20_kbd is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.              *
 *********************************************************************/

/*********************************************************************
 * Hardware:
 *  
 *  A o---o D7 --+
 *  B o---o D6   |
 *  C o---o D5   |
 *  D o---o D4   |
 *  E o---o D3   |
 *  F o---o D2   |
 *  G o---o D1   |
 *  H o---o D0   |
 *  0 o---o A0   |
 *  1 o---o A1   |
 *  2 o---o D13  |
 *  3 o---o D12  |
 *  4 o---o D11  |
 *  5 o---o D10  |
 *  6 o---o D9   |
 *  7 o---o D8   |
 *  8 o----------+
 *  
 *  I o---o A2
 * 
 * Power LED to A3 over a 1K resistor
 *
 *********************************************************************/

#include <avr/pgmspace.h>

//#define SERIAL_DEBUG 1

#define LEONARDO_OFFSET 136
#define MODMASK_LEFT_CTRL   0b00000001
#define MODMASK_LEFT_SHIFT  0b00000010
#define MODMASK_LEFT_ALT    0b00000100
#define MODMASK_LEFT_GUI    0b00001000
#define MODMASK_RIGHT_CTRL  0b00010000
#define MODMASK_RIGHT_SHIFT 0b00100000
#define MODMASK_RIGHT_ALT   0b01000000
#define MODMASK_RIGHT_GUI   0b10000000
/* Number of rows in keyboard matrix */
#define NUMROWS 9

/*
 * The USB keycodes are enumerated here - this is simply an
 * enumeration of the allowed scan-codes used for USB HID devices
 */
enum keycodes {
  HKEY__= LEONARDO_OFFSET,
  HKEY_errorRollOver,
  HKEY_POSTfail,
  HKEY_errorUndefined,
  HKEY_A,        // 4
  HKEY_B,
  HKEY_C,
  HKEY_D,
  HKEY_E,
  HKEY_F,
  HKEY_G, 
  HKEY_H,
  HKEY_I,
  HKEY_J,
  HKEY_K,
  HKEY_L,
  HKEY_M,        // 0x10
  HKEY_N,
  HKEY_O,
  HKEY_P,
  HKEY_Q, 
  HKEY_R,
  HKEY_S,
  HKEY_T,
  HKEY_U,
  HKEY_V,
  HKEY_W,
  HKEY_X,
  HKEY_Y,
  HKEY_Z,
  HKEY_1,
  HKEY_2,
  HKEY_3,        // 0x20
  HKEY_4,
  HKEY_5,
  HKEY_6,
  HKEY_7,
  HKEY_8,
  HKEY_9,
  HKEY_0,        // 0x27
  HKEY_enter,
  HKEY_esc,
  HKEY_bckspc,   // backspace
  HKEY_tab,
  HKEY_spc,      // space
  HKEY_minus,    // - (and _)
  HKEY_equal,    // = (and +)
  HKEY_lbr,      // [
  HKEY_rbr,      // ]  -- 0x30
  HKEY_bckslsh,  // \ (and |)
  HKEY_hash,     // Non-US # and ~
  HKEY_smcol,    // ; (and :)
  HKEY_ping,     // ' and "
  HKEY_grave,    // Grave accent and tilde
  HKEY_comma,    // , (and <)
  HKEY_dot,      // . (and >)
  HKEY_slash,    // / (and ?)
  HKEY_cpslck,   // capslock
  HKEY_F1,
  HKEY_F2,
  HKEY_F3,
  HKEY_F4,
  HKEY_F5,
  HKEY_F6, 
  HKEY_F7,       // 0x40
  HKEY_F8,
  HKEY_F9,
  HKEY_F10,
  HKEY_F11,
  HKEY_F12,
  HKEY_PrtScr,
  HKEY_scrlck,
  HKEY_break,
  HKEY_ins,
  HKEY_home,
  HKEY_pgup,
  HKEY_del,
  HKEY_end,
  HKEY_pgdn,
  HKEY_rarr, 
  HKEY_larr,     // 0x50
  HKEY_darr,
  HKEY_uarr,
  HKEY_numlock,
  HKEY_KPslash,
  HKEY_KPast,
  HKEY_KPminus,
  HKEY_KPplus,
  HKEY_KPenter,
  HKEY_KP1,
  HKEY_KP2,
  HKEY_KP3,
  HKEY_KP4,
  HKEY_KP5,
  HKEY_KP6,
  HKEY_KP7,
  HKEY_KP8,      // 0x60
  HKEY_KP9,
  HKEY_KP0,
  HKEY_KPcomma,
  HKEY_Euro2
};

/*********************************************************************
 * Keyboard Mappings                                                 *
 * Just uncomment one of them. dk and us are not tested              *
 *********************************************************************/

#include "key_de.h"
//#include "key_dk.h"
//#include "key_us.h"

/* ******** Definitions ***************** */

const int ledPin       = A3;
const int joyLeftRight = A4;
const int joyFwBckFire = A5;
	
/* ******** Main Body ******************* */

/* This array holds the last values of the scanned keyboard matrix */
static unsigned char debounceBuf[NUMROWS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
/* This array holds the values of the keyboard matrix which have been allready sent to the host computer */
static unsigned char scannedKeys[NUMROWS] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

void setup(){
#ifdef SERIAL_DEBUG
  Serial.begin(9600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for Leonardo only
  }
#else
  Keyboard.begin();
#endif

  // Initialize the key matrix column pins for input with pullup resistor
  pinMode(7, INPUT); digitalWrite(7, HIGH);  // Col A and I
  pinMode(6, INPUT); digitalWrite(6, HIGH);  // Col B
  pinMode(5, INPUT); digitalWrite(5, HIGH);  // Col C
  pinMode(4, INPUT); digitalWrite(4, HIGH);  // Col D
  pinMode(3, INPUT); digitalWrite(3, HIGH);  // Col E
  pinMode(2, INPUT); digitalWrite(2, HIGH);  // Col F
  pinMode(1, INPUT); digitalWrite(1, HIGH);  // Col G
  pinMode(0, INPUT); digitalWrite(0, HIGH);  // Col H

  // Initialize the key matrix row pins for input without pullup resistor
  pinMode(8,  INPUT); digitalWrite(8,  LOW);  // Row 0
  pinMode(9,  INPUT); digitalWrite(9,  LOW);  // Row 1
  pinMode(10, INPUT); digitalWrite(10, LOW);  // Row 2
  pinMode(11, INPUT); digitalWrite(11, LOW);  // Row 3
  pinMode(12, INPUT); digitalWrite(12, LOW);  // Row 4
  pinMode(13, INPUT); digitalWrite(13, LOW);  // Row 5
  pinMode(A0, INPUT); digitalWrite(A0, LOW);  // Row 6
  pinMode(A1, INPUT); digitalWrite(A1, LOW);  // Row 7
  pinMode(A2, INPUT); digitalWrite(A2, LOW);  // Row 8

  // Turn ON the Power LED
  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, HIGH);
  
#ifdef SERIAL_DEBUG
  Serial.println("Init done");
#endif
}

void selectRow(int row) {
  pinMode(8,  INPUT);
  pinMode(9,  INPUT);
  pinMode(10, INPUT);
  pinMode(11, INPUT);
  pinMode(12, INPUT);
  pinMode(13, INPUT);
  pinMode(A0, INPUT);
  pinMode(A1, INPUT);
  pinMode(A2, INPUT);
  switch (row) {
    case 0: pinMode(A0, OUTPUT); break;
    case 1: pinMode(A1, OUTPUT); break;
    case 2: pinMode(13, OUTPUT); break;
    case 3: pinMode(12, OUTPUT); break;
    case 4: pinMode(11, OUTPUT); break;
    case 5: pinMode(10, OUTPUT); break;
    case 6: pinMode(9,  OUTPUT); break;
    case 7: pinMode(8,  OUTPUT); break;
    case 8: pinMode(A2, OUTPUT); break;
    default: ;
  }
}

uint8_t readCols() {
  uint8_t cols = 0;
  
  cols |= (~PINE & 0b01000000) >> 6;  // Bit 0    1
  cols |= (~PIND & 0b10000000) >> 6;  // Bit 1    2
  cols |= (~PINC & 0b01000000) >> 4;  // Bit 2    4
  cols |= (~PIND & 0b00010000) >> 1;  // Bit 3    8
  cols |= (~PIND & 0b00000001) << 4;  // Bit 4   16
  cols |= (~PIND & 0b00000010) << 4;  // Bit 5   32
  cols |= (~PIND & 0b00001000) << 3;  // Bit 6   64
  cols |= (~PIND & 0b00000100) << 5;  // Bit 7  128
  return cols;
}

/* lrbfF000 */
uint8_t readJoystick() {
  uint8_t joy = 0;
  int analog = analogRead(joyLeftRight);
  
  if (analog < 489) {  // right
    joy |= 0b01000000;
  } else if (analog < 839) {  // left
    joy |= 0b10000000;
  }
  analog = analogRead(joyFwBckFire);
  if (analog < 311) {  // forward + fire
    joy |= 0b00011000;
  } else if (analog < 436) {  // forward
    joy |= 0b00010000;
  } else if (analog < 603) {  // back + fire
    joy |= 0b00101000;
  } else if (analog < 722) {  // back
    joy |= 0b00100000;
  } else if (analog < 905) {  // fire
    joy |= 0b00001000;
  }
  return joy;
}

/*
 * Converts hid keys to ascii keys suitable for serial output
 */
uint8_t hid2ascii(uint8_t key, uint8_t modifiers)
{
  if (key < 128) return key;  // it's an ascii key, not a hid key
  if (key < 136) return 0;  // it's a modifier
  if (key >= HKEY_A && key <= HKEY_Z) {
    if (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) {
      return key - HKEY_A + 'A';
    } 
    else {
      return key - HKEY_A + 'a';
    }
  }
  else if (key >= HKEY_1 && key <= HKEY_9) {
    return key - HKEY_1 + '1';
  }
  else if (key >= HKEY_KP1 && key <= HKEY_KP9) {
    return key - HKEY_KP1 + '1';
  }
  else {
    switch (key) {
      case HKEY_0:       return '0';
      case HKEY_enter:   return 0x0A;
      case HKEY_bckspc:  return 0x08;
      case HKEY_tab:     return 0x09;
      case HKEY_spc:     return ' ';
      case HKEY_minus:   return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '_' : '-';
      case HKEY_equal:   return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '+' : '=';
      case HKEY_lbr:     return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '{' : '[';
      case HKEY_rbr:     return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '}' : ']';
      case HKEY_bckslsh: return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '|' : '\\';
      case HKEY_hash:    return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '~' : '#';
      case HKEY_smcol:   return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? ':' : ';';
      case HKEY_ping:    return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '"' : '\'';
      case HKEY_grave:   return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '~' : '`';
      case HKEY_comma:   return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '<' : ',';
      case HKEY_dot:     return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '>' : '.';
      case HKEY_slash:   return (modifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) ? '?' : '/';
      case HKEY_KPslash: return '/';
      case HKEY_KPast:   return '*';
      case HKEY_KPminus: return '-';
      case HKEY_KPplus:  return '+';
      case HKEY_KPenter: return 0x0A;
      case HKEY_KP0:     return '0';
      case HKEY_KPcomma: return ',';
      default: return 0;
    }
  }
}

void mod2Serial(uint8_t key, boolean pressed) {
#ifdef SERIAL_DEBUG
  if (key < KEY_LEFT_CTRL || key > KEY_RIGHT_GUI) return;
  if (pressed) Serial.write("v");
  else Serial.write("^");
  switch(key) {
    case KEY_LEFT_CTRL: Serial.print("LC"); break;
    case KEY_LEFT_SHIFT: Serial.print("LS"); break;
    case KEY_LEFT_ALT: Serial.print("LA"); break;
    case KEY_LEFT_GUI: Serial.print("LG"); break;
    case KEY_RIGHT_CTRL: Serial.print("RC"); break;
    case KEY_RIGHT_SHIFT: Serial.print("RS"); break;
    case KEY_RIGHT_ALT: Serial.print("RA"); break;
    case KEY_RIGHT_GUI: Serial.print("RG");
  } 
#endif
}

/*
 * returns the new modifiers mask
 */
uint8_t hidReport(uint8_t key, uint8_t modifiers, boolean pressed) {
  static uint8_t lastModifiers = 0;
  uint8_t changedMods;
  uint8_t cnt, mask;
  
  // set/reset modifiers from key
  if (key >= KEY_LEFT_CTRL && key <= KEY_RIGHT_GUI) {
    if (pressed) modifiers |= (1<<(key-128));
    else modifiers &= ~(1<<(key-128));
    key = 0;
  }
  // report changed modifiers to Keyboard library
  changedMods = modifiers ^ lastModifiers;
  for (cnt=0, mask=1; cnt<8; cnt++, mask<<=1) {
    if (changedMods & mask) {
      uint8_t modKey = KEY_LEFT_CTRL + cnt;
      
#ifdef SERIAL_DEBUG
      mod2Serial(modKey, modifiers & mask);
#else
      if (modifiers & mask) Keyboard.press(modKey);
      else Keyboard.release(modKey);
#endif
    }
  }
  lastModifiers = modifiers;
  // report key to Keyboard library
  if (key) {
#ifdef SERIAL_DEBUG
    if (pressed) Serial.write(hid2ascii(key, modifiers));
#else
    if (pressed) Keyboard.press(key);
    else Keyboard.release(key);
#endif
  }
  return modifiers;
}

uint8_t hidReleaseAll() {
  hidReport(0, 0, false);
  Keyboard.releaseAll();
  return 0;
}

/* This function scans the entire keyboard, debounces the keys, and
   if a key change has been found, a new report is generated, and the
   function returns true to signal the transfer of the report. */
unsigned char scanKeys(void) {
  static uint8_t curModifiers = 0;
  static uint8_t debounceCnt = 10;
  uint8_t retval = 0;
  uint8_t data, row, key, modkeys;
  uint8_t col, mask;

  for (row=0; row<NUMROWS; row++) { /* Scan all rows */
    selectRow(row);
    delayMicroseconds(30);
    data = readCols();
    if (row == 8) data |= readJoystick();
    if (data ^ debounceBuf[row]) {
      debounceCnt = 10; /* If a change was detected, activate debounce counter */
    }
    debounceBuf[row] = data; /* Store the result */
  }

  if (debounceCnt == 1) { /* Debounce counter expired */
    data = 0;
    for (row=0; row<NUMROWS; row++) { /* If all data is null, call the releaseAll Function */
      data |= debounceBuf[row];
    }
    if (!data) curModifiers = hidReleaseAll();
    modkeys = 0;
    for (row=0; row<NUMROWS; row++) { /* Process all rows for key-codes */
      unsigned char sdata = scannedKeys[row];
      unsigned char smask;

      data  = debounceBuf[row]; /* Restore buffer */
      smask = data ^ sdata;      
      if (smask != 0x00) { /* Anything changed on this row? - optimization */
        for (col=0, mask=1; col<8; col++, mask<<=1) { /* yes - check individual bits */
          if (smask & mask) { /* Change detected */
            uint8_t clearMods = 0;
            
            key = pgm_read_byte_near(keymap + row*8 + col); /* Read keyboard map */            
            if (key > 0 && key < 128) { /* Special handling of shifted keys */
              uint8_t keynum    = key - 1;
              
              if (curModifiers & (MODMASK_LEFT_SHIFT | MODMASK_RIGHT_SHIFT)) {
                key     = pgm_read_byte_near(spec_keys + keynum*4 + 2); /* Shifted */
                modkeys = pgm_read_byte_near(spec_keys + keynum*4 + 3);
              } else {
                key     = pgm_read_byte_near(spec_keys + keynum*4); /* Unmodified */
                modkeys = pgm_read_byte_near(spec_keys + keynum*4 + 1);
                /* forced shift without pressed shift key should be undone after keypress */
                if (modkeys == MODMASK_LEFT_SHIFT) clearMods = modkeys;
              }
              if (modkeys & 0x80) { /* Clear RSHIFT */
                curModifiers &= ~0x20;
              }
              if (modkeys & 0x08) { /* Clear LSHIFT */
                curModifiers &= ~0x02;
              }
              curModifiers |= modkeys & 0x77; /* Set other modifiers */
            }
            if (key) { /* Normal keycode should be added to report */
              curModifiers = hidReport(key, curModifiers, mask & data);
            }
            if (clearMods) curModifiers &= ~(clearMods);           
          }  // end change detect
        }
        scannedKeys[row] = data;
      }  // end anything changed
    }
    retval|=1; /* Must have been a change at some point, since debounce is done */
  }
  if (debounceCnt) debounceCnt--;
  return retval;
}

void loop(){
  while(1) {
    scanKeys();
  }
}


